from pymysql.err import *

def map_umysql_error_to_umysqldb_exception(umysql_exc):
    errorclass = error_map.get(umysql_exc.args[0])
    if errorclass:
        return errorclass(*umysql_exc.args)

    # couldn't find the right error number
    code, message  = umysql_exc.args[:2]
    if code == 0 and message.startswith("Connection reset by peer"):
        return OperationalError(2013, "Lost connection to MySQL server during query")

    return InternalError(*umysql_exc.args)

def map_runtime_error_to_umysqldb_exception(exc):
    if exc.args == ('Not connected',):
        return ProgrammingError("cursor closed")
    return exc



import sys
import pymysql
from pymysql import *
from pymysql import DATETIME
from pymysql import __all__

from .util import setdocstring

__all__ += ['DATETIME']

@setdocstring(pymysql.Connect)
def Connect(*args, **kwargs):
    from .connections import Connection
    return Connection(*args, **kwargs)

def thread_safe():
    return True  # match MySQLdb.thread_safe()

def install_as_MySQLdb():
    """
    After this function is called, any application that imports MySQLdb or
    _mysql will unwittingly actually use umysqldb.
    """
    sys.modules["MySQLdb"] = sys.modules["_mysql"] = sys.modules["umysqldb"]
    sys.modules["MySQLdb.constants"] = sys.modules["pymysql.constants"]


connect = Connection = Connect



import sys
import pymysql.cursors

from .util import setdocstring

class Cursor(pymysql.cursors.Cursor):
    @setdocstring(pymysql.cursors.Cursor.execute)
    def execute(self, query, args=None):
        if args is None:
            args = ()
        elif not isinstance(args, (tuple, list, dict)):
            args = (args,)

        result = 0
        try:
            result = self._query(query, args)
        except:
            exc, value, tb = sys.exc_info()
            del tb
            self.errorhandler(self, exc, value)

        self._executed = query
        return result

    @setdocstring(pymysql.cursors.Cursor.executemany)
    def executemany(self, query, args):
        if not args:
            return
        self.rowcount = sum([self.execute(query, arg) for arg in args])
        return self.rowcount

    def _query(self, query, args=()):
        conn = self._get_db()
        conn.query(query, args)
        self.rowcount = conn._result.affected_rows
        self.rownumber = 0
        self.description = conn._result.description
        self.lastrowid = conn._result.insert_id
        self._rows = conn._result.rows
        self._has_next = 0
        return self.rowcount


class DictCursor(Cursor):
    """A cursor which returns results as a dictionary"""

    def execute(self, query, args=None):
        result = super(DictCursor, self).execute(query, args)
        if self.description:
            self._fields = [ field[0] for field in self.description ]
        return result

    def fetchone(self):
        ''' Fetch the next row '''
        self._check_executed()
        if self._rows is None or self.rownumber >= len(self._rows):
            return None
        result = dict(zip(self._fields, self._rows[self.rownumber]))
        self.rownumber += 1
        return result

    def fetchmany(self, size=None):
        ''' Fetch several rows '''
        self._check_executed()
        if self._rows is None:
            return None
        end = self.rownumber + (size or self.arraysize)
        result = [ dict(zip(self._fields, r)) for r in self._rows[self.rownumber:end] ]
        self.rownumber = min(end, len(self._rows))
        return tuple(result)

    def fetchall(self):
        ''' Fetch all the rows '''
        self._check_executed()
        if self._rows is None:
            return None
        if self.rownumber:
            result = [ dict(zip(self._fields, r)) for r in self._rows[self.rownumber:] ]
        else:
            result = [ dict(zip(self._fields, r)) for r in self._rows ]
        self.rownumber = len(self._rows)
        return tuple(result)




import sys
import time
import datetime
import socket

import umysql
import pymysql.connections
from pymysql.constants import FIELD_TYPE

from .util import setdocstring
from .cursors import Cursor
from .err import (
    map_umysql_error_to_umysqldb_exception,
    map_runtime_error_to_umysqldb_exception,
    Error,
    OperationalError,
)
from .times import (
    encode_struct_time,
    encode_timedelta,
    encode_time,
    TimeDelta_or_None,
    mysql_timestamp_converter,
)


encoders = {
    time.struct_time : encode_struct_time,
    datetime.timedelta: encode_timedelta,
    datetime.time: encode_time,
}

decoders = {
    FIELD_TYPE.TIME: TimeDelta_or_None,
    FIELD_TYPE.TIMESTAMP: mysql_timestamp_converter,
    FIELD_TYPE.NEWDECIMAL: float,
}

def notouch(x):
    return x

def defaulterrorhandler(connection, cursor, errorclass, errorvalue):
    raise errorclass, errorvalue


class ResultSet(object):
    def __init__(self, affected_rows=None, insert_id=None, description=None,
                 rows=None):
        self.affected_rows = affected_rows
        self.insert_id = insert_id
        self.description = description
        self.rows = rows


class Connection(pymysql.connections.Connection):

    """MySQL Database Connection Object"""

    errorhandler = defaulterrorhandler

    @setdocstring(pymysql.connections.Connection.__init__)
    def __init__(self, *args, **kwargs):
        if 'cursorclass' not in kwargs:
            kwargs['cursorclass'] = Cursor
        if 'conv' not in kwargs:
            kwargs['conv'] = decoders
        if 'charset' not in kwargs:
            kwargs['charset'] = 'utf8'
        self._umysql_conn = umysql.Connection()
        super(Connection, self).__init__(*args, **kwargs)

    @setdocstring(pymysql.connections.Connection.set_charset)
    def set_charset(self, charset):
        if charset:
            self._umysql_conn.query("SET NAMES %s", (charset,))
            self.charset = charset

    @setdocstring(pymysql.connections.Connection.autocommit)
    def autocommit(self, value):
        self._umysql_conn.query("SET AUTOCOMMIT = %s", (value,))

    @setdocstring(pymysql.connections.Connection.commit)
    def commit(self):
        self.query('COMMIT')

    @setdocstring(pymysql.connections.Connection.rollback)
    def rollback(self):
        self.query("ROLLBACK")

    @setdocstring(pymysql.connections.Connection.close)
    def close(self):
        if not self._umysql_conn.is_connected():
            raise Error("Already closed")
        self._umysql_conn.close()


    def _connect(self):
        try:
            self._umysql_conn.connect(self.host, self.port, self.user,
                                      self.password, self.db or '', False, self.charset)
        except socket.error, e:
            raise OperationalError(2003, "Can't connect to MySQL server on %r (%s)" % (
                self.host, e.args[0]))
        except umysql.Error, exc:
            traceback = sys.exc_info()[2]
            exc = map_umysql_error_to_umysqldb_exception(exc)
            raise exc, None, traceback

    # internal use only (called from cursor)
    def query(self, sql, args=()):
        args = self._convert_args(args)
        try:
            result_set = self._umysql_conn.query(sql, args)
        except umysql.Error, exc:
            traceback = sys.exc_info()[2]
            exc = map_umysql_error_to_umysqldb_exception(exc)
            raise exc, None, traceback
        except RuntimeError, exc:
            traceback = sys.exc_info()[2]
            exc = map_runtime_error_to_umysqldb_exception(exc)
            raise exc, None, traceback
        else:
            self._result = self._convert_result_set(result_set)
            return self._result.affected_rows

    def _convert_args(self, args):
        args = tuple(encoders.get(type(arg), notouch)(arg)
                     for arg in args)
        return args

    def _convert_result_set(self, result_set):
        if isinstance(result_set, tuple):
            rs = ResultSet(affected_rows=result_set[0],
                           insert_id=result_set[1])
        else:
            converters = [self.decoders.get(field[1]) for field in
                          result_set.fields]
            rows = tuple(tuple(conv(data) if conv and data is not None else data
                          for data, conv in zip(row, converters))
                         for row in result_set.rows)
            description = tuple(f + (None,) * (7 - len(f)) for f in result_set.fields)
            rs = ResultSet(description=description, rows=rows,
                           affected_rows=len(rows))
        return rs

    # _mysql support
    def get_proto_info(self):
        raise NotImplementedError("umysql has no proto info")

    def get_server_info(self):
        raise NotImplementedError("umysql has no server info")

    def thread_id(self):
        raise NotImplementedError("umysql has no thread info")

    def ping(self, reconnect=True):
        raise NotImplementedError("umysql has no ping support")




import time
from datetime import datetime, date, timedelta
import math

Timestamp = datetime

def encode_struct_time(obj):
    return time.strftime("%Y-%m-%d %H:%M:%S", obj)

def encode_timedelta(delta):
    seconds = int(delta.seconds) % 60
    minutes = int(delta.seconds // 60) % 60
    hours = int(delta.seconds // 3600) % 24 + int(delta.days) * 24
    return "%02d:%02d:%02d" % (hours, minutes, seconds)

def encode_time(obj):
    s = "%02d:%02d:%02d" % (int(obj.hour), int(obj.minute), int(obj.second))
    if obj.microsecond:
        s += ".%f" % obj.microsecond
    return s

def Date_or_None(s):
    try: return date(*[ int(x) for x in s.split('-',2)])
    except: return None


def DateTime_or_None(s):
    if ' ' in s:
        sep = ' '
    elif 'T' in s:
        sep = 'T'
    else:
        return Date_or_None(s)

    try:
        d, t = s.split(sep, 1)
        return datetime(*[ int(x) for x in d.split('-')+t.split(':') ])
    except:
        return Date_or_None(s)

def TimeDelta_or_None(s):
    try:
        h, m, s = s.split(':')
        h, m, s = int(h), int(m), float(s)
        td = timedelta(hours=abs(h), minutes=m, seconds=int(s),
                       microseconds=int(math.modf(s)[0] * 1000000))
        if h < 0:
            return -td
        else:
            return td
    except ValueError:
        # unpacking or int/float conversion failed
        return None

def mysql_timestamp_converter(s):
    """Convert a MySQL TIMESTAMP to a Timestamp object."""
    # MySQL>4.1 returns TIMESTAMP in the same format as DATETIME
    if s[4] == '-': return DateTime_or_None(s)
    s += "0"*(14-len(s))
    parts = [int(x) for x in (s[:4],s[4:6],s[6:8],s[8:10],s[10:12],s[12:14])
             if x]
    try:
        return Timestamp(*parts)
    except Exception:
        return None
